# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

pyarrow = pytest.importorskip("pyarrow")


def test_toarrow_BitMaskedArray():
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    bitmask = ak.layout.IndexU8(np.array([40, 34], dtype=np.uint8))
    array = ak.layout.BitMaskedArray(bitmask, content, False, 9, False)
    assert ak.to_arrow(array).to_pylist() == ak.to_list(array)


def test_toarrow_ByteMaskedArray_1():
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    bytemask = ak.layout.Index8(np.array([False, True, False], dtype=np.bool))
    array = ak.layout.ByteMaskedArray(bytemask, content, True)
    assert ak.to_arrow(array).to_pylist() == ak.to_list(array)


def test_toarrow_NumpyArray_1():
    array = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5]))
    assert isinstance(ak.to_arrow(array), (pyarrow.lib.Tensor, pyarrow.lib.Array))
    assert ak.to_arrow(array).to_pylist() == [0.0, 1.1, 2.2, 3.3, 4.4, 5.5]


def test_toarrow_NumpyArray_2():
    array = ak.layout.NumpyArray(np.array([[0.0, 1.1], [2.2, 3.3], [4.4, 5.5]]))
    assert isinstance(ak.to_arrow(array), (pyarrow.lib.Tensor, pyarrow.lib.Array))
    assert ak.to_arrow(array) == pyarrow.Tensor.from_numpy(
        np.array([[0.0, 1.1], [2.2, 3.3], [4.4, 5.5]])
    )


def test_toarrow_EmptyArray():
    array = ak.layout.EmptyArray()
    assert isinstance(ak.to_arrow(array), (pyarrow.lib.Array))
    assert ak.to_arrow(array).to_pylist() == []


def test_toarrow_ListOffsetArray64():
    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.layout.ListOffsetArray64(offsets, content)
    assert isinstance(ak.to_arrow(array), (pyarrow.ListArray))
    assert ak.to_arrow(array).to_pylist() == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]


def test_toarrow_ListOffsetArrayU32():
    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.IndexU32(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.layout.ListOffsetArrayU32(offsets, content)
    assert isinstance(ak.to_arrow(array), (pyarrow.ListArray))
    assert ak.to_arrow(array).to_pylist() == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]


def test_toarrow_ListArray_RegularArray():
    # Testing parameters
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    offsets = ak.layout.Index32(np.array([0, 3, 3, 5, 6, 9]))
    array = ak.layout.ListOffsetArray32(offsets, content)
    assert ak.to_arrow(array).to_pylist() == [
        ["one", "two", "three"],
        [],
        ["four", "five"],
        ["six"],
        ["seven", "eight", "nine"],
    ]

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    regulararray = ak.layout.RegularArray(listoffsetarray, 2)
    starts = ak.layout.Index64(np.array([0, 1]))
    stops = ak.layout.Index64(np.array([2, 3]))
    listarray = ak.layout.ListArray64(starts, stops, regulararray)

    assert isinstance(ak.to_arrow(listarray), (pyarrow.ListArray))
    assert ak.to_arrow(listarray).to_pylist() == [
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4], [5.5]]],
        [[[3.3, 4.4], [5.5]], [[6.6, 7.7, 8.8, 9.9], []]],
    ]

    assert isinstance(ak.to_arrow(regulararray), (pyarrow.ListArray))
    assert ak.to_arrow(regulararray).to_pylist() == [
        [[0.0, 1.1, 2.2], []],
        [[3.3, 4.4], [5.5]],
        [[6.6, 7.7, 8.8, 9.9], []],
    ]


def test_toarrow_RecordArray():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5]))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index32(np.array([0, 3, 3, 5, 6, 9]))

    recordarray = ak.layout.RecordArray(
        [content1, listoffsetarray, content2, content1],
        keys=["one", "two", "2", "wonky"],
    )

    assert isinstance(ak.to_arrow(recordarray), (pyarrow.StructArray))
    assert ak.to_arrow(recordarray).to_pylist() == [
        {"one": 1, "two": [0.0, 1.1, 2.2], "2": 1.1, "wonky": 1},
        {"one": 2, "two": [], "2": 2.2, "wonky": 2},
        {"one": 3, "two": [3.3, 4.4], "2": 3.3, "wonky": 3},
        {"one": 4, "two": [5.5], "2": 4.4, "wonky": 4},
        {"one": 5, "two": [6.6, 7.7, 8.8, 9.9], "2": 5.5, "wonky": 5},
    ]


def test_toarrow_UnionArray():
    content0 = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout
    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5]))
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.int8))
    index = ak.layout.Index32(np.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=np.int32))
    unionarray = ak.layout.UnionArray8_32(tags, index, [content0, content1])

    assert isinstance(ak.to_arrow(unionarray), (pyarrow.UnionArray))
    assert ak.to_arrow(unionarray).to_pylist() == [
        1,
        2,
        [1.1, 2.2, 3.3],
        [],
        3,
        [4.4, 5.5],
        5,
        4,
    ]


def test_toarrow_IndexedArray():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index = ak.layout.Index32(np.array([0, 2, 4, 6, 8, 9, 7, 5], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray32(index, content)
    do_this_instead = ak.to_categorical(indexedarray, highlevel=False)

    assert isinstance(ak.to_arrow(indexedarray), pyarrow.lib.DoubleArray)
    assert ak.to_arrow(indexedarray).to_pylist() == [
        0.0,
        2.2,
        4.4,
        6.6,
        8.8,
        9.9,
        7.7,
        5.5,
    ]

    assert isinstance(ak.to_arrow(do_this_instead), pyarrow.DictionaryArray)
    assert ak.to_arrow(do_this_instead).to_pylist() == [
        0.0,
        2.2,
        4.4,
        6.6,
        8.8,
        9.9,
        7.7,
        5.5,
    ]


def test_toarrow_IndexedOptionArray_2():
    array = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5, None])

    assert ak.to_arrow(array).to_pylist() == [1.1, 2.2, 3.3, 4.4, 5.5, None]
    assert ak.to_arrow(array[:-1]).to_pylist() == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert ak.to_arrow(array[:1]).to_pylist() == [1.1]
    assert ak.to_arrow(array[:0]).to_pylist() == []

    content = ak.layout.NumpyArray(np.array([], dtype=np.float64))
    index = ak.layout.Index32(np.array([-1, -1, -1, -1], dtype=np.int32))
    indexedoptionarray = ak.layout.IndexedOptionArray32(index, content)
    assert ak.to_arrow(indexedoptionarray).to_pylist() == [None, None, None, None]


def test_toarrow_ByteMaskedArray_2():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, True, False, False, False], dtype=np.int8)),
        listoffsetarray,
        True,
    )

    assert ak.to_arrow(bytemaskedarray).to_pylist() == [
        [0.0, 1.1, 2.2],
        [],
        None,
        None,
        None,
    ]


def test_toarrow_ByteMaskedArray_3():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    regulararray = ak.layout.RegularArray(listoffsetarray, 2)
    starts = ak.layout.Index64(np.array([0, 1]))
    stops = ak.layout.Index64(np.array([2, 3]))
    listarray = ak.layout.ListArray64(starts, stops, regulararray)

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, False], dtype=np.int8)), listarray, True
    )
    assert ak.to_arrow(bytemaskedarray).to_pylist() == ak.to_list(bytemaskedarray)


def test_toarrow_ByteMaskedArray_4():
    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5]))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    recordarray = ak.layout.RecordArray(
        [content1, listoffsetarray, content2, content1],
        keys=["one", "two", "2", "wonky"],
    )

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, False], dtype=np.int8)), recordarray, True
    )
    assert ak.to_arrow(bytemaskedarray).to_pylist() == ak.to_list(bytemaskedarray)


def test_toarrow_ByteMaskedArray_5():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index = ak.layout.Index32(np.array([0, 2, 4, 6, 8, 9, 7, 5], dtype=np.int64))
    indexedarray = ak.layout.IndexedArray32(index, content)

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, False, False], dtype=np.int8)),
        indexedarray,
        True,
    )
    assert ak.to_arrow(bytemaskedarray).to_pylist() == ak.to_list(bytemaskedarray)


def test_toarrow_ByteMaskedArray_broken_unions_1():
    content0 = ak.Array(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    ).layout
    content1 = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1, 0, 0], dtype=np.int8))
    index = ak.layout.Index32(np.array([0, 1, 1, 0, 2, 2, 4, 3, 3, 4], dtype=np.int32))
    unionarray = ak.layout.UnionArray8_32(tags, index, [content0, content1])

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(
            # tags          1,     1,     0,    0,     1,    0,    1,     1,    0,     0
            # index         0,     1,     1,    0,     2,    2,    4,     3,    3,     4
            np.array(
                [True, False, False, True, False, True, True, False, False, True],
                dtype=np.int8,
            )
        ),
        unionarray,
        valid_when=True,
    )
    assert ak.to_arrow(bytemaskedarray).to_pylist() == ak.to_list(bytemaskedarray)


def test_toarrow_ByteMaskedArray_broken_unions_2():
    content0 = ak.Array(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    ).layout
    content1 = ak.layout.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4]))
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1, 0, 0, 0], dtype=np.int8))
    index = ak.layout.Index32(
        np.array([0, 1, 1, 0, 2, 2, 4, 3, 3, 4, 3], dtype=np.int32)
    )
    unionarray = ak.layout.UnionArray8_32(tags, index, [content0, content1])

    bytemaskedarray = ak.layout.ByteMaskedArray(
        ak.layout.Index8(
            # tags          1,     1,     0,    0,     1,    0,    1,     1,    0,     0,    0
            # index         0,     1,     1,    0,     2,    2,    4,     3,    3,     4,    3
            np.array(
                [True, False, False, True, False, True, True, False, False, True, True],
                dtype=np.int8,
            )
        ),
        unionarray,
        valid_when=True,
    )
    assert ak.to_arrow(bytemaskedarray).to_pylist() == ak.to_list(bytemaskedarray)


def test_toarrow_IndexedOptionArray():
    ioa = ak.layout.IndexedOptionArray32(
        ak.layout.Index32([-30, 19, 6, 7, -3, 21, 13, 22, 17, 9, -12, 16]),
        ak.layout.NumpyArray(
            np.array(
                [
                    5.2,
                    1.7,
                    6.7,
                    -0.4,
                    4.0,
                    7.8,
                    3.8,
                    6.8,
                    4.2,
                    0.3,
                    4.6,
                    6.2,
                    6.9,
                    -0.7,
                    3.9,
                    1.6,
                    8.7,
                    -0.7,
                    3.2,
                    4.3,
                    4.0,
                    5.8,
                    4.2,
                    7.0,
                    5.6,
                    3.8,
                ]
            )
        ),
    )
    assert ak.to_arrow(ioa).to_pylist() == ak.to_list(ioa)


def test_fromarrow_NumpyArray_1():
    boolarray = ak.layout.NumpyArray(
        np.array([True, True, True, False, False, True, False, True, False, True])
    )
    assert ak.to_list(
        ak.from_arrow(ak.to_arrow(boolarray), highlevel=False)
    ) == ak.to_list(boolarray)


def test_fromarrow_NumpyArray_2():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    assert ak.to_list(
        ak.from_arrow(ak.to_arrow(content), highlevel=False)
    ) == ak.to_list(content)


def test_fromarrow_ListOffsetArray():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)
    assert ak.to_list(
        ak.from_arrow(ak.to_arrow(listoffsetarray), highlevel=False)
    ) == ak.to_list(listoffsetarray)


def test_fromarrow_RegularArray():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)

    regulararray = ak.layout.RegularArray(listoffsetarray, 2)
    assert ak.to_list(
        ak.from_arrow(ak.to_arrow(regulararray), highlevel=False)
    ) == ak.to_list(regulararray)


def test_fromarrow_RecordArray():
    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.10])
    )
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10, 10]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, content)

    content1 = ak.layout.NumpyArray(np.array([1, 2, 3, 4, 5]))
    content2 = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.layout.Index32(np.array([0, 3, 3, 5, 6, 9]))
    recordarray = ak.layout.RecordArray(
        [content1, listoffsetarray, content2, content1],
        keys=["one", "chonks", "2", "wonky"],
    )
    assert ak.to_list(
        ak.from_arrow(ak.to_arrow(recordarray), highlevel=False)
    ) == ak.to_list(recordarray)


def test_fromarrow_UnionArray():
    content0 = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]).layout
    content = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight", "nine"]
    ).layout
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.int8))
    index = ak.layout.Index32(np.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=np.int32))
    array = ak.layout.UnionArray8_32(tags, index, [content0, content])
    assert ak.to_list(ak.from_arrow(ak.to_arrow(array), highlevel=False)) == ak.to_list(
        array
    )


def test_chunkedarray():
    a = pyarrow.chunked_array(
        [
            pyarrow.array([1.1, 2.2, 3.3]),
            pyarrow.array([], pyarrow.float64()),
            pyarrow.array([4.4, 5.5]),
            pyarrow.array([6.6]),
            pyarrow.array([], pyarrow.float64()),
            pyarrow.array([], pyarrow.float64()),
            pyarrow.array([7.7, 8.8, 9.9]),
        ]
    )
    assert a.to_pylist() == [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        7.7,
        8.8,
        9.9,
    ]


def test_recordbatch():
    a = pyarrow.RecordBatch.from_arrays(
        [
            pyarrow.array([1.1, 2.2, 3.3, 4.4, 5.5]),
            pyarrow.array([[1, 2, 3], [], [], [4, 5], [6]]),
        ],
        ["a", "b"],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"a": 1.1, "b": [1, 2, 3]},
        {"a": 2.2, "b": []},
        {"a": 3.3, "b": []},
        {"a": 4.4, "b": [4, 5]},
        {"a": 5.5, "b": [6]},
    ]

    a = pyarrow.RecordBatch.from_arrays(
        [
            pyarrow.array([1.1, 2.2, 3.3, None, 5.5]),
            pyarrow.array([[1, None, 3], [], [], [4, 5], [6]]),
        ],
        ["a", "b"],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"a": 1.1, "b": [1, None, 3]},
        {"a": 2.2, "b": []},
        {"a": 3.3, "b": []},
        {"a": None, "b": [4, 5]},
        {"a": 5.5, "b": [6]},
    ]

    a = pyarrow.RecordBatch.from_arrays(
        [
            pyarrow.array([1.1, 2.2, 3.3, None, 5.5]),
            pyarrow.array([[1, 2, 3], [], [4, 5], [None], [6]]),
            pyarrow.array(
                [
                    {"x": 1, "y": 1.1},
                    {"x": 2, "y": 2.2},
                    {"x": 3, "y": 3.3},
                    {"x": 4, "y": None},
                    {"x": 5, "y": 5.5},
                ]
            ),
            pyarrow.array(
                [
                    {"x": 1, "y": 1.1},
                    None,
                    None,
                    {"x": 4, "y": None},
                    {"x": 5, "y": 5.5},
                ]
            ),
            pyarrow.array(
                [
                    [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
                    [],
                    [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
                    [None],
                    [{"x": 6, "y": 6.6}],
                ]
            ),
        ],
        ["a", "b", "c", "d", "e"],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {
            "a": 1.1,
            "b": [1, 2, 3],
            "c": {"x": 1, "y": 1.1},
            "d": {"x": 1, "y": 1.1},
            "e": [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        },
        {"a": 2.2, "b": [], "c": {"x": 2, "y": 2.2}, "d": None, "e": []},
        {
            "a": 3.3,
            "b": [4, 5],
            "c": {"x": 3, "y": 3.3},
            "d": None,
            "e": [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
        },
        {
            "a": None,
            "b": [None],
            "c": {"x": 4, "y": None},
            "d": {"x": 4, "y": None},
            "e": [None],
        },
        {
            "a": 5.5,
            "b": [6],
            "c": {"x": 5, "y": 5.5},
            "d": {"x": 5, "y": 5.5},
            "e": [{"x": 6, "y": 6.6}],
        },
    ]


### All of the following tests were copied (translated) over from Awkward 0.


def test_arrow_toarrow_string():
    a = ak.from_iter(["one", "two", "three"], highlevel=False)
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a), highlevel=False)) == ak.to_list(a)
    a = ak.from_iter([["one", "two", "three"], [], ["four", "five"]], highlevel=False)
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a), highlevel=False)) == ak.to_list(a)
    if hasattr(pyarrow.BinaryArray, "from_buffers"):
        a = ak.from_iter([b"one", b"two", b"three"], highlevel=False)
        assert ak.to_list(ak.from_arrow(ak.to_arrow(a), highlevel=False)) == [
            b"one",
            b"two",
            b"three",
        ]
        a = ak.from_iter(
            [[b"one", b"two", b"three"], [], [b"four", b"five"]], highlevel=False
        )
        assert ak.to_list(ak.from_arrow(ak.to_arrow(a), highlevel=False)) == [
            [b"one", b"two", b"three"],
            [],
            [b"four", b"five"],
        ]
    else:
        a = ak.from_iter([b"one", b"two", b"three"], highlevel=False)
        assert ak.to_list(ak.from_arrow(ak.to_arrow(a), highlevel=False)) == [
            "one",
            "two",
            "three",
        ]
        a = ak.from_iter(
            [[b"one", b"two", b"three"], [], [b"four", b"five"]], highlevel=False
        )
        assert ak.to_list(ak.from_arrow(ak.to_arrow(a), highlevel=False)) == [
            ["one", "two", "three"],
            [],
            ["four", "five"],
        ]


def test_arrow_array():
    a = pyarrow.array([1.1, 2.2, 3.3, 4.4, 5.5])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [1.1, 2.2, 3.3, 4.4, 5.5]


def test_arrow_boolean():
    a = pyarrow.array([True, True, False, False, True])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        True,
        True,
        False,
        False,
        True,
    ]


def test_arrow_array_null():
    a = pyarrow.array([1.1, 2.2, 3.3, None, 4.4, 5.5])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        1.1,
        2.2,
        3.3,
        None,
        4.4,
        5.5,
    ]


def test_arrow_nested_array():
    a = pyarrow.array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]


def test_arrow_nested_nested_array():
    a = pyarrow.array([[[1.1, 2.2], [3.3], []], [], [[4.4, 5.5]]])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [[1.1, 2.2], [3.3], []],
        [],
        [[4.4, 5.5]],
    ]


def test_arrow_nested_array_null():
    a = pyarrow.array([[1.1, 2.2, None], [], [4.4, 5.5]])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [1.1, 2.2, None],
        [],
        [4.4, 5.5],
    ]


def test_arrow_null_nested_array_null():
    a = pyarrow.array([[1.1, 2.2, None], [], None, [4.4, 5.5]])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [1.1, 2.2, None],
        [],
        None,
        [4.4, 5.5],
    ]


def test_arrow_chunked_array():
    a = pyarrow.chunked_array(
        [
            pyarrow.array([1.1, 2.2, 3.3, 4.4, 5.5]),
            pyarrow.array([], pyarrow.float64()),
            pyarrow.array([6.6, 7.7, 8.8]),
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
        6.6,
        7.7,
        8.8,
    ]


def test_arrow_struct():
    a = pyarrow.array([{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"x": 1, "y": 1.1},
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
    ]


def test_arrow_struct_null():
    a = pyarrow.array([{"x": 1, "y": 1.1}, {"x": 2, "y": None}, {"x": 3, "y": 3.3}])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"x": 1, "y": 1.1},
        {"x": 2, "y": None},
        {"x": 3, "y": 3.3},
    ]


def test_arrow_null_struct():
    a = pyarrow.array(
        [{"x": 1, "y": 1.1}, None, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"x": 1, "y": 1.1},
        None,
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
    ]


def test_arrow_null_struct_null():
    a = pyarrow.array(
        [{"x": 1, "y": 1.1}, None, {"x": 2, "y": None}, {"x": 3, "y": 3.3}]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"x": 1, "y": 1.1},
        None,
        {"x": 2, "y": None},
        {"x": 3, "y": 3.3},
    ]


def test_arrow_chunked_struct():
    t = pyarrow.struct({"x": pyarrow.int64(), "y": pyarrow.float64()})
    a = pyarrow.chunked_array(
        [
            pyarrow.array(
                [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}], t
            ),
            pyarrow.array([], t),
            pyarrow.array([{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}], t),
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"x": 1, "y": 1.1},
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
        {"x": 4, "y": 4.4},
        {"x": 5, "y": 5.5},
    ]


def test_arrow_nested_struct():
    a = pyarrow.array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
            [],
            [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        [],
        [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
    ]


def test_arrow_nested_struct_null():
    a = pyarrow.array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": None}, {"x": 3, "y": 3.3}],
            [],
            [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": None}, {"x": 3, "y": 3.3}],
        [],
        [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
    ]


def test_arrow_null_nested_struct():
    a = pyarrow.array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
            None,
            [],
            [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        None,
        [],
        [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
    ]


def test_arrow_null_nested_struct_null():
    a = pyarrow.array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": None}, {"x": 3, "y": 3.3}],
            None,
            [],
            [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": None}, {"x": 3, "y": 3.3}],
        None,
        [],
        [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
    ]


def test_arrow_struct_nested():
    a = pyarrow.array(
        [{"x": [], "y": 1.1}, {"x": [2], "y": 2.2}, {"x": [3, 3], "y": 3.3}]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"x": [], "y": 1.1},
        {"x": [2], "y": 2.2},
        {"x": [3, 3], "y": 3.3},
    ]


def test_arrow_struct_nested_null():
    a = pyarrow.array(
        [{"x": [], "y": 1.1}, {"x": [2], "y": 2.2}, {"x": [None, 3], "y": 3.3}]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {"x": [], "y": 1.1},
        {"x": [2], "y": 2.2},
        {"x": [None, 3], "y": 3.3},
    ]


def test_arrow_nested_struct_nested():
    a = pyarrow.array(
        [
            [{"x": [], "y": 1.1}, {"x": [2], "y": 2.2}, {"x": [3, 3], "y": 3.3}],
            [],
            [{"x": [4, 4, 4], "y": 4.4}, {"x": [5, 5, 5, 5], "y": 5.5}],
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [{"x": [], "y": 1.1}, {"x": [2], "y": 2.2}, {"x": [3, 3], "y": 3.3}],
        [],
        [{"x": [4, 4, 4], "y": 4.4}, {"x": [5, 5, 5, 5], "y": 5.5}],
    ]


def test_arrow_null_nested_struct_nested_null():
    a = pyarrow.array(
        [
            [{"x": [], "y": 1.1}, {"x": [2], "y": 2.2}, {"x": [None, 3], "y": 3.3}],
            None,
            [],
            [{"x": [4, 4, 4], "y": 4.4}, {"x": [5, 5, 5, 5], "y": 5.5}],
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        [{"x": [], "y": 1.1}, {"x": [2], "y": 2.2}, {"x": [None, 3], "y": 3.3}],
        None,
        [],
        [{"x": [4, 4, 4], "y": 4.4}, {"x": [5, 5, 5, 5], "y": 5.5}],
    ]


def test_arrow_strings():
    if not ak._util.py27:
        a = pyarrow.array(["one", "two", "three", u"fo\u2014ur", "five"])
        assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
            "one",
            "two",
            "three",
            u"fo\u2014ur",
            "five",
        ]


def test_arrow_strings_null():
    if not ak._util.py27:
        a = pyarrow.array(["one", "two", None, u"fo\u2014ur", "five"])
        assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
            "one",
            "two",
            None,
            u"fo\u2014ur",
            "five",
        ]


def test_arrow_binary():
    a = pyarrow.array([b"one", b"two", b"three", b"four", b"five"])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        b"one",
        b"two",
        b"three",
        b"four",
        b"five",
    ]


def test_arrow_binary_null():
    a = pyarrow.array([b"one", b"two", None, b"four", b"five"])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        b"one",
        b"two",
        None,
        b"four",
        b"five",
    ]


def test_arrow_chunked_strings():
    a = pyarrow.chunked_array(
        [
            pyarrow.array(["one", "two", "three", "four", "five"]),
            pyarrow.array(["six", "seven", "eight"]),
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
    ]


def test_arrow_nested_strings():
    a = pyarrow.array([["one", "two", "three"], [], ["four", "five"]])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        ["one", "two", "three"],
        [],
        ["four", "five"],
    ]


def test_arrow_nested_strings_null():
    a = pyarrow.array([["one", "two", None], [], ["four", "five"]])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        ["one", "two", None],
        [],
        ["four", "five"],
    ]


def test_arrow_null_nested_strings_null():
    a = pyarrow.array([["one", "two", None], [], None, ["four", "five"]])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        ["one", "two", None],
        [],
        None,
        ["four", "five"],
    ]


def test_arrow_union_sparse():
    a = pyarrow.UnionArray.from_sparse(
        pyarrow.array([0, 1, 0, 0, 1], type=pyarrow.int8()),
        [
            pyarrow.array([0.0, 1.1, 2.2, 3.3, 4.4]),
            pyarrow.array([True, True, False, True, False]),
        ],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [0.0, True, 2.2, 3.3, False]


def test_arrow_union_sparse_null():
    a = pyarrow.UnionArray.from_sparse(
        pyarrow.array([0, 1, 0, 0, 1], type=pyarrow.int8()),
        [
            pyarrow.array([0.0, 1.1, None, 3.3, 4.4]),
            pyarrow.array([True, True, False, True, False]),
        ],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        0.0,
        True,
        None,
        3.3,
        False,
    ]


def test_arrow_union_sparse_null_null():
    a = pyarrow.UnionArray.from_sparse(
        pyarrow.array([0, 1, 0, 0, 1], type=pyarrow.int8()),
        [
            pyarrow.array([0.0, 1.1, None, 3.3, 4.4]),
            pyarrow.array([True, None, False, True, False]),
        ],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        0.0,
        None,
        None,
        3.3,
        False,
    ]


def test_arrow_union_dense():
    a = pyarrow.UnionArray.from_dense(
        pyarrow.array([0, 1, 0, 0, 0, 1, 1], type=pyarrow.int8()),
        pyarrow.array([0, 0, 1, 2, 3, 1, 2], type=pyarrow.int32()),
        [pyarrow.array([0.0, 1.1, 2.2, 3.3]), pyarrow.array([True, True, False])],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        0.0,
        True,
        1.1,
        2.2,
        3.3,
        True,
        False,
    ]


def test_arrow_union_dense_null():
    a = pyarrow.UnionArray.from_dense(
        pyarrow.array([0, 1, 0, 0, 0, 1, 1], type=pyarrow.int8()),
        pyarrow.array([0, 0, 1, 2, 3, 1, 2], type=pyarrow.int32()),
        [pyarrow.array([0.0, 1.1, None, 3.3]), pyarrow.array([True, True, False])],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        0.0,
        True,
        1.1,
        None,
        3.3,
        True,
        False,
    ]


def test_arrow_union_dense_null_null():
    a = pyarrow.UnionArray.from_dense(
        pyarrow.array([0, 1, 0, 0, 0, 1, 1], type=pyarrow.int8()),
        pyarrow.array([0, 0, 1, 2, 3, 1, 2], type=pyarrow.int32()),
        [pyarrow.array([0.0, 1.1, None, 3.3]), pyarrow.array([True, None, False])],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        0.0,
        True,
        1.1,
        None,
        3.3,
        None,
        False,
    ]


def test_arrow_dictarray():
    a = pyarrow.DictionaryArray.from_arrays(
        pyarrow.array([0, 0, 2, 2, 1, 0, 2, 1, 1]),
        pyarrow.array(["one", "two", "three"]),
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        "one",
        "one",
        "three",
        "three",
        "two",
        "one",
        "three",
        "two",
        "two",
    ]


def test_arrow_dictarray_null():
    a = pyarrow.DictionaryArray.from_arrays(
        pyarrow.array([0, 0, 2, None, 1, None, 2, 1, 1]),
        pyarrow.array(["one", "two", "three"]),
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        "one",
        "one",
        "three",
        None,
        "two",
        None,
        "three",
        "two",
        "two",
    ]


def test_arrow_null_dictarray():
    a = pyarrow.DictionaryArray.from_arrays(
        pyarrow.array([0, 0, 2, 2, 1, 0, 2, 1, 1]),
        pyarrow.array(["one", None, "three"]),
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        "one",
        "one",
        "three",
        "three",
        None,
        "one",
        "three",
        None,
        None,
    ]


def test_arrow_batch():
    a = pyarrow.RecordBatch.from_arrays(
        [
            pyarrow.array([1.1, 2.2, 3.3, None, 5.5]),
            pyarrow.array([[1, 2, 3], [], [4, 5], [None], [6]]),
            pyarrow.array(
                [
                    {"x": 1, "y": 1.1},
                    {"x": 2, "y": 2.2},
                    {"x": 3, "y": 3.3},
                    {"x": 4, "y": None},
                    {"x": 5, "y": 5.5},
                ]
            ),
            pyarrow.array(
                [
                    {"x": 1, "y": 1.1},
                    None,
                    None,
                    {"x": 4, "y": None},
                    {"x": 5, "y": 5.5},
                ]
            ),
            pyarrow.array(
                [
                    [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
                    [],
                    [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
                    [None],
                    [{"x": 6, "y": 6.6}],
                ]
            ),
        ],
        ["a", "b", "c", "d", "e"],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {
            "a": 1.1,
            "b": [1, 2, 3],
            "c": {"x": 1, "y": 1.1},
            "d": {"x": 1, "y": 1.1},
            "e": [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        },
        {"a": 2.2, "b": [], "c": {"x": 2, "y": 2.2}, "d": None, "e": []},
        {
            "a": 3.3,
            "b": [4, 5],
            "c": {"x": 3, "y": 3.3},
            "d": None,
            "e": [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
        },
        {
            "a": None,
            "b": [None],
            "c": {"x": 4, "y": None},
            "d": {"x": 4, "y": None},
            "e": [None],
        },
        {
            "a": 5.5,
            "b": [6],
            "c": {"x": 5, "y": 5.5},
            "d": {"x": 5, "y": 5.5},
            "e": [{"x": 6, "y": 6.6}],
        },
    ]


def test_arrow_table():
    a = pyarrow.Table.from_batches(
        [
            pyarrow.RecordBatch.from_arrays(
                [
                    pyarrow.array([1.1, 2.2, 3.3, None, 5.5]),
                    pyarrow.array([[1, 2, 3], [], [4, 5], [None], [6]]),
                    pyarrow.array(
                        [
                            {"x": 1, "y": 1.1},
                            {"x": 2, "y": 2.2},
                            {"x": 3, "y": 3.3},
                            {"x": 4, "y": None},
                            {"x": 5, "y": 5.5},
                        ]
                    ),
                    pyarrow.array(
                        [
                            {"x": 1, "y": 1.1},
                            None,
                            None,
                            {"x": 4, "y": None},
                            {"x": 5, "y": 5.5},
                        ]
                    ),
                    pyarrow.array(
                        [
                            [
                                {"x": 1, "y": 1.1},
                                {"x": 2, "y": 2.2},
                                {"x": 3, "y": 3.3},
                            ],
                            [],
                            [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
                            [None],
                            [{"x": 6, "y": 6.6}],
                        ]
                    ),
                ],
                ["a", "b", "c", "d", "e"],
            ),
            pyarrow.RecordBatch.from_arrays(
                [
                    pyarrow.array([1.1, 2.2, 3.3, None, 5.5]),
                    pyarrow.array([[1, 2, 3], [], [4, 5], [None], [6]]),
                    pyarrow.array(
                        [
                            {"x": 1, "y": 1.1},
                            {"x": 2, "y": 2.2},
                            {"x": 3, "y": 3.3},
                            {"x": 4, "y": None},
                            {"x": 5, "y": 5.5},
                        ]
                    ),
                    pyarrow.array(
                        [
                            {"x": 1, "y": 1.1},
                            None,
                            None,
                            {"x": 4, "y": None},
                            {"x": 5, "y": 5.5},
                        ]
                    ),
                    pyarrow.array(
                        [
                            [
                                {"x": 1, "y": 1.1},
                                {"x": 2, "y": 2.2},
                                {"x": 3, "y": 3.3},
                            ],
                            [],
                            [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
                            [None],
                            [{"x": 6, "y": 6.6}],
                        ]
                    ),
                ],
                ["a", "b", "c", "d", "e"],
            ),
        ]
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [
        {
            "a": 1.1,
            "b": [1, 2, 3],
            "c": {"x": 1, "y": 1.1},
            "d": {"x": 1, "y": 1.1},
            "e": [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        },
        {"a": 2.2, "b": [], "c": {"x": 2, "y": 2.2}, "d": None, "e": []},
        {
            "a": 3.3,
            "b": [4, 5],
            "c": {"x": 3, "y": 3.3},
            "d": None,
            "e": [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
        },
        {
            "a": None,
            "b": [None],
            "c": {"x": 4, "y": None},
            "d": {"x": 4, "y": None},
            "e": [None],
        },
        {
            "a": 5.5,
            "b": [6],
            "c": {"x": 5, "y": 5.5},
            "d": {"x": 5, "y": 5.5},
            "e": [{"x": 6, "y": 6.6}],
        },
        {
            "a": 1.1,
            "b": [1, 2, 3],
            "c": {"x": 1, "y": 1.1},
            "d": {"x": 1, "y": 1.1},
            "e": [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        },
        {"a": 2.2, "b": [], "c": {"x": 2, "y": 2.2}, "d": None, "e": []},
        {
            "a": 3.3,
            "b": [4, 5],
            "c": {"x": 3, "y": 3.3},
            "d": None,
            "e": [{"x": 4, "y": None}, {"x": 5, "y": 5.5}],
        },
        {
            "a": None,
            "b": [None],
            "c": {"x": 4, "y": None},
            "d": {"x": 4, "y": None},
            "e": [None],
        },
        {
            "a": 5.5,
            "b": [6],
            "c": {"x": 5, "y": 5.5},
            "d": {"x": 5, "y": 5.5},
            "e": [{"x": 6, "y": 6.6}],
        },
    ]


def test_arrow_nonnullable_table():
    x = pyarrow.array([1, 2, 3])
    y = pyarrow.array([1.1, 2.2, 3.3])
    table = pyarrow.Table.from_arrays([x], ["x"])
    if hasattr(pyarrow, "column"):
        table2 = table.add_column(
            1,
            pyarrow.column(
                pyarrow.field("y", y.type, False), np.array([1.1, 2.2, 3.3])
            ),
        )
    else:
        table2 = table.add_column(1, "y", y)
    assert ak.to_list(ak.from_arrow(table2, highlevel=False)) == [
        {"x": 1, "y": 1.1},
        {"x": 2, "y": 2.2},
        {"x": 3, "y": 3.3},
    ]


def test_arrow_coverage100():
    a = ak.from_iter(
        [True, True, False, False, True, False, True, False], highlevel=False
    )
    assert ak.to_arrow(a).to_pylist() == ak.to_list(a)

    a = ak.layout.ListOffsetArray32(
        ak.layout.Index32(np.array([0, 5, 10], "i4")),
        ak.layout.NumpyArray(
            np.frombuffer(b"hellothere", "u1"), parameters={"__array__": "bytes"}
        ),
        parameters={"__array__": "bytestring"},
    )
    assert ak.to_arrow(a).to_pylist() == [b"hello", b"there"]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True, False, False, True, True])),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10, 15, 20, 25, 30], "i4")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellotherehellotherehellothere", "u1"),
                parameters={"__array__": "bytes"},
            ),
            parameters={"__array__": "bytestring"},
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == [
        b"hello",
        None,
        b"hello",
        b"there",
        None,
        None,
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True])),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10], "i4")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellothere", "u1"), parameters={"__array__": "bytes"}
            ),
            parameters={"__array__": "bytestring"},
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == [b"hello", None]

    a = ak.layout.IndexedOptionArray32(
        ak.layout.Index32(np.array([-1, 1, -1, 0, 0, -1], "i4")),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10], "i4")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellothere", "u1"), parameters={"__array__": "bytes"}
            ),
            parameters={"__array__": "bytestring"},
        ),
    )
    assert ak.to_arrow(a).to_pylist() == [
        None,
        b"there",
        None,
        b"hello",
        b"hello",
        None,
    ]

    a = ak.layout.ListOffsetArray32(
        ak.layout.Index32(np.array([0, 5, 10], "i4")),
        ak.layout.NumpyArray(
            np.frombuffer(b"hellothere", "u1"), parameters={"__array__": "chars"}
        ),
        parameters={"__array__": "string"},
    )
    assert ak.to_arrow(a).to_pylist() == ["hello", "there"]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True, False, False, True, True])),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10, 15, 20, 25, 30], "i4")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellotherehellotherehellothere", "u1"),
                parameters={"__array__": "chars"},
            ),
            parameters={"__array__": "string"},
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == ["hello", None, "hello", "there", None, None]
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a))) == [
        "hello",
        None,
        "hello",
        "there",
        None,
        None,
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True, False, False, True, True])),
        ak.layout.ListOffsetArray64(
            ak.layout.Index64(np.array([0, 5, 10, 15, 20, 25, 30], "i8")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellotherehellotherehellothere", "u1"),
                parameters={"__array__": "chars"},
            ),
            parameters={"__array__": "string"},
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == ["hello", None, "hello", "there", None, None]
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a))) == [
        "hello",
        None,
        "hello",
        "there",
        None,
        None,
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True, False, False, True, True])),
        ak.layout.ListOffsetArray64(
            ak.layout.Index64(np.array([0, 5, 10, 15, 20, 25, 30], "i8")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellotherehellotherehellothere", "u1"),
                parameters={"__array__": "bytes"},
            ),
            parameters={"__array__": "bytestring"},
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == [
        b"hello",
        None,
        b"hello",
        b"there",
        None,
        None,
    ]
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a))) == [
        b"hello",
        None,
        b"hello",
        b"there",
        None,
        None,
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True])),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10], "i4")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellothere", "u1"), parameters={"__array__": "chars"}
            ),
            parameters={"__array__": "string"},
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == ["hello", None]

    a = ak.layout.IndexedOptionArray32(
        ak.layout.Index32(np.array([-1, 1, -1, 0, 0, -1], "i4")),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10], "i4")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellothere", "u1"), parameters={"__array__": "chars"}
            ),
            parameters={"__array__": "string"},
        ),
    )
    assert ak.to_arrow(a).to_pylist() == [None, "there", None, "hello", "hello", None]

    a = ak.layout.ListOffsetArray32(
        ak.layout.Index32(np.array([0, 5, 10], "i4")),
        ak.layout.NumpyArray(np.frombuffer(b"hellothere", "u1")),
    )
    assert ak.to_arrow(a).to_pylist() == [
        [104, 101, 108, 108, 111],
        [116, 104, 101, 114, 101],
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True, False, False, True, True])),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10, 15, 20, 25, 30], "i4")),
            ak.layout.NumpyArray(
                np.frombuffer(b"hellotherehellotherehellothere", "u1")
            ),
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == [
        [104, 101, 108, 108, 111],
        None,
        [104, 101, 108, 108, 111],
        [116, 104, 101, 114, 101],
        None,
        None,
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, True])),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10], "i4")),
            ak.layout.NumpyArray(np.frombuffer(b"hellothere", "u1")),
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == [[104, 101, 108, 108, 111], None]

    a = ak.layout.IndexedOptionArray32(
        ak.layout.Index32(np.array([-1, 1, -1, 0, 0, -1], "i4")),
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10], "i4")),
            ak.layout.NumpyArray(np.frombuffer(b"hellothere", "u1")),
        ),
    )
    assert ak.to_arrow(a).to_pylist() == [
        None,
        [116, 104, 101, 114, 101],
        None,
        [104, 101, 108, 108, 111],
        [104, 101, 108, 108, 111],
        None,
    ]

    a = ak.layout.IndexedOptionArray32(
        ak.layout.Index32(np.array([-1, 1, -1, 0, 0, -1], "i4")),
        ak.layout.RegularArray(
            ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])), 3
        ),
    )
    assert ak.to_arrow(a).to_pylist() == [
        None,
        [4.4, 5.5, 6.6],
        None,
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        None,
    ]

    a = ak.layout.IndexedOptionArray32(
        ak.layout.Index32(np.array([-1, 1, -1, 0, 0, -1, 1, -1], "i4")),
        ak.layout.RegularArray(
            ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])), 3
        ),
    )
    assert ak.to_arrow(a).to_pylist() == [
        None,
        [4.4, 5.5, 6.6],
        None,
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        None,
        [4.4, 5.5, 6.6],
        None,
    ]

    a = ak.layout.IndexedOptionArray64(
        ak.layout.Index64(np.array([-1, 1, -1, 0, 0, -1, 1, -1], "i8")),
        ak.layout.RegularArray(
            ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])), 3
        ),
    )
    assert ak.to_arrow(a).to_pylist() == [
        None,
        [4.4, 5.5, 6.6],
        None,
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        None,
        [4.4, 5.5, 6.6],
        None,
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, True, True, True, False, False])),
        ak.layout.IndexedOptionArray32(
            ak.layout.Index32(np.array([-1, 1, -1, 0, 0, -1], "i4")),
            ak.layout.RegularArray(
                ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])), 3
            ),
        ),
        valid_when=True,
    )
    assert ak.to_arrow(a).to_pylist() == [
        None,
        [4.4, 5.5, 6.6],
        None,
        [1.1, 2.2, 3.3],
        None,
        None,
    ]

    a = ak.layout.UnmaskedArray(
        ak.layout.ListOffsetArray32(
            ak.layout.Index32(np.array([0, 5, 10], "i4")),
            ak.layout.NumpyArray(np.frombuffer(b"hellothere", "u1")),
        )
    )
    assert ak.to_arrow(a).to_pylist() == [
        [104, 101, 108, 108, 111],
        [116, 104, 101, 114, 101],
    ]

    a = pyarrow.array(
        ["one", "two", "three", "two", "two", "one", "three", "one"]
    ).dictionary_encode()
    b = ak.from_arrow(a, highlevel=False)
    assert isinstance(b, ak._util.indexedtypes)
    assert ak.to_list(b) == ["one", "two", "three", "two", "two", "one", "three", "one"]

    a = ak.Array([[1.1, 2.2, 3.3], [], None, [4.4, 5.5]])
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a))) == [
        [1.1, 2.2, 3.3],
        [],
        None,
        [4.4, 5.5],
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, False, False, True, True, False, False])),
        ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 999, 314, 4.4, 5.5])),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == [1.1, 2.2, 3.3, None, None, 4.4, 5.5]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([False, False, False, True, True, False, False])),
        ak.from_iter(
            [b"hello", b"", b"there", b"yuk", b"", b"o", b"hellothere"], highlevel=False
        ),
        valid_when=False,
    )
    assert ak.to_arrow(a).to_pylist() == [
        b"hello",
        b"",
        b"there",
        None,
        None,
        b"o",
        b"hellothere",
    ]

    a = ak.layout.ByteMaskedArray(
        ak.layout.Index8([True, True, False, True]),
        ak.from_iter([[1.1, 2.2, 3.3], [], [999], [4.4, 5.5]], highlevel=False),
        valid_when=True,
    )
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a))) == [
        [1.1, 2.2, 3.3],
        [],
        None,
        [4.4, 5.5],
    ]

    a = ak.from_iter([[1, 2, 3], [], [4, 5], 999, 123], highlevel=False)
    assert ak.to_arrow(a).to_pylist() == [[1, 2, 3], [], [4, 5], 999, 123]
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a))) == [
        [1, 2, 3],
        [],
        [4, 5],
        999,
        123,
    ]


def test_arrow_coverage100_broken_unions():
    a = ak.from_iter([[1, 2, 3], [], [4, 5], 999, 123], highlevel=False)
    b = ak.layout.ByteMaskedArray(
        ak.layout.Index8(np.array([True, True, False, False, True])), a, valid_when=True
    )
    assert ak.to_arrow(b).to_pylist() == [[1, 2, 3], [], None, None, 123]
    assert ak.to_list(ak.from_arrow(ak.to_arrow(b))) == [[1, 2, 3], [], None, None, 123]

    content1 = ak.from_iter([1.1, 2.2, 3.3, 4.4, 5.5], highlevel=False)
    content2 = ak.layout.NumpyArray(np.array([], dtype=np.int32))
    a = ak.layout.UnionArray8_32(
        ak.layout.Index8(np.array([0, 0, 0, 0, 0], "i1")),
        ak.layout.Index32(np.array([0, 1, 2, 3, 4], "i4")),
        [content1, content2],
    )
    assert ak.to_list(a) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert ak.to_arrow(a).to_pylist() == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert ak.to_list(ak.from_arrow(ak.to_arrow(a))) == [1.1, 2.2, 3.3, 4.4, 5.5]

    a = pyarrow.UnionArray.from_sparse(
        pyarrow.array([0, 0, 0, 0, 0], type=pyarrow.int8()),
        [
            pyarrow.array([0.0, 1.1, None, 3.3, 4.4]),
            pyarrow.array([True, None, False, True, False]),
        ],
    )
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [0.0, 1.1, None, 3.3, 4.4]

    a = pyarrow.UnionArray.from_sparse(
        pyarrow.array([0, 1, 0, 1, 1], "i1"),
        [
            pyarrow.array([[0.0, 1.1, 2.2], [], None, [5.5], [6.6, 7.7, 8.8, 9.9]]),
            pyarrow.array([0.0, 1.1, 2.2, None, None]),
        ],
        ["0", "1"],
        [0, 1],
    )
    assert a.to_pylist() == [[0.0, 1.1, 2.2], 1.1, None, None, None]
    assert ak.to_list(ak.from_arrow(a)) == [[0.0, 1.1, 2.2], 1.1, None, None, None]

    a = pyarrow.chunked_array([pyarrow.array([1.1, 2.2, 3.3, 4.4, 5.5])])
    assert ak.to_list(ak.from_arrow(a, highlevel=False)) == [1.1, 2.2, 3.3, 4.4, 5.5]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import json

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    example = ak.Array([[{"x": 1, "y": []}, {"x": 2, "y": [1, 1]}], []])
    cache = {}
    virtualarray = ak.virtual(
        lambda: example, form=example.layout.form, length=len(example)
    )
    assert len(cache) == 0

    tmp1 = virtualarray["x"]
    assert json.loads(str(tmp1.layout.form)) == {
        "class": "VirtualArray",
        "form": {"class": "ListOffsetArray64", "offsets": "i64", "content": "int64"},
        "has_length": True,
    }
    assert len(cache) == 0

    tmp2 = virtualarray["y"]
    assert json.loads(str(tmp2.layout.form)) == {
        "class": "VirtualArray",
        "form": {
            "class": "ListOffsetArray64",
            "offsets": "i64",
            "content": {
                "class": "ListOffsetArray64",
                "offsets": "i64",
                "content": "int64",
            },
        },
        "has_length": True,
    }
    assert len(cache) == 0

    assert tmp1.tolist() == [[1, 2], []]
    assert tmp2.tolist() == [[[], [1, 1]], []]


def test_no_break_regular_broadcasting():
    # because we redefined Form::has_virtual_form to agree with an interpretation of its name in English

    assert (np.array([[1, 2, 3], [4, 5, 6]]) + np.array([[10], [20]])).tolist() == [
        [11, 12, 13],
        [24, 25, 26],
    ]
    assert (
        ak.Array(np.array([[1, 2, 3], [4, 5, 6]])) + ak.Array(np.array([[10], [20]]))
    ).tolist() == [[11, 12, 13], [24, 25, 26]]
    with pytest.raises(ValueError):
        ak.Array([[1, 2, 3], [4, 5, 6]]) + ak.Array([[10], [20]])
    left, right = (
        ak.Array(np.array([[1, 2, 3], [4, 5, 6]])),
        ak.Array(np.array([[10], [20]])),
    )
    assert (
        ak.virtual(lambda: left, form=left.layout.form, length=len(left))
        + ak.virtual(lambda: right, form=right.layout.form, length=len(right))
    ).tolist() == [[11, 12, 13], [24, 25, 26]]
    with pytest.raises(ValueError):
        left, right = ak.Array([[1, 2, 3], [4, 5, 6]]), ak.Array([[10], [20]])
        ak.virtual(lambda: left, form=left.layout.form, length=len(left)) + ak.virtual(
            lambda: right, form=right.layout.form, length=len(right)
        )

    assert (np.array([[1, 2, 3], [4, 5, 6]]) + np.array([10])).tolist() == [
        [11, 12, 13],
        [14, 15, 16],
    ]
    assert (
        ak.Array(np.array([[1, 2, 3], [4, 5, 6]])) + ak.Array(np.array([10]))
    ).tolist() == [[11, 12, 13], [14, 15, 16]]
    assert (ak.Array([[1, 2, 3], [4, 5, 6]]) + ak.Array([10])).tolist() == [
        [11, 12, 13],
        [14, 15, 16],
    ]
    left, right = ak.Array(np.array([[1, 2, 3], [4, 5, 6]])), ak.Array(np.array([10]))
    assert (
        ak.virtual(lambda: left, form=left.layout.form, length=len(left))
        + ak.virtual(lambda: right, form=right.layout.form, length=len(right))
    ).tolist() == [[11, 12, 13], [14, 15, 16]]
    left, right = ak.Array([[1, 2, 3], [4, 5, 6]]), ak.Array([10])
    assert (
        ak.virtual(lambda: left, form=left.layout.form, length=len(left))
        + ak.virtual(lambda: right, form=right.layout.form, length=len(right))
    ).tolist() == [[11, 12, 13], [14, 15, 16]]

    assert (ak.Array([[1, 2, 3], [4, 5, 6]]) + ak.Array([10, 20])).tolist() == [
        [11, 12, 13],
        [24, 25, 26],
    ]
    left, right = ak.Array([[1, 2, 3], [4, 5, 6]]), ak.Array([10, 20])
    assert (
        ak.virtual(lambda: left, form=left.layout.form, length=len(left))
        + ak.virtual(lambda: right, form=right.layout.form, length=len(right))
    ).tolist() == [[11, 12, 13], [24, 25, 26]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


numba = pytest.importorskip("numba")


def test():
    @numba.njit
    def do_something(array):
        out = np.zeros(len(array), np.bool_)
        for i, x in enumerate(array):
            if x:
                out[i] = x
        return out

    array = ak.Array([True, False, False])
    assert do_something(array).tolist() == [True, False, False]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import warnings
import sys
import distutils.version

import awkward as ak

checked_version = False


def import_numexpr():
    global checked_version
    try:
        import numexpr
    except ImportError:
        raise ImportError(
            """install the 'numexpr' package with:

    pip install numexpr --upgrade

or

    conda install numexpr"""
        )
    else:
        if not checked_version and distutils.version.LooseVersion(
            numexpr.__version__
        ) < distutils.version.LooseVersion("2.7.1"):
            warnings.warn(
                "Awkward Array is only known to work with numexpr 2.7.1 or later"
                "(you have version {0})".format(numexpr.__version__),
                RuntimeWarning,
            )
        checked_version = True
        return numexpr


def getArguments(names, local_dict=None, global_dict=None):
    call_frame = sys._getframe(2)

    clear_local_dict = False
    if local_dict is None:
        local_dict = call_frame.f_locals
        clear_local_dict = True
    try:
        frame_globals = call_frame.f_globals
        if global_dict is None:
            global_dict = frame_globals

        clear_local_dict = clear_local_dict and frame_globals is not local_dict

        arguments = []
        for name in names:
            try:
                a = local_dict[name]
            except KeyError:
                a = global_dict[name]
            arguments.append(a)  # <--- different from NumExpr
    finally:
        if clear_local_dict:
            local_dict.clear()

    return arguments


def evaluate(
    expression, local_dict=None, global_dict=None, order="K", casting="safe", **kwargs
):
    numexpr = import_numexpr()

    context = numexpr.necompiler.getContext(kwargs, frame_depth=1)
    expr_key = (expression, tuple(sorted(context.items())))
    if expr_key not in numexpr.necompiler._names_cache:
        numexpr.necompiler._names_cache[expr_key] = numexpr.necompiler.getExprNames(
            expression, context
        )
    names, ex_uses_vml = numexpr.necompiler._names_cache[expr_key]
    arguments = getArguments(names, local_dict, global_dict)

    arrays = [
        ak.operations.convert.to_layout(x, allow_record=True, allow_other=True)
        for x in arguments
    ]

    def getfunction(inputs, depth):
        if all(
            isinstance(x, ak.layout.NumpyArray) or not isinstance(x, ak.layout.Content)
            for x in inputs
        ):
            return lambda: (
                ak.layout.NumpyArray(
                    numexpr.evaluate(
                        expression,
                        dict(zip(names, inputs)),
                        {},
                        order=order,
                        casting=casting,
                        **kwargs
                    )
                ),
            )
        else:
            return None

    behavior = ak._util.behaviorof(*arrays)
    out = ak._util.broadcast_and_apply(
        arrays, getfunction, behavior, allow_records=False
    )
    assert isinstance(out, tuple) and len(out) == 1
    return ak._util.wrap(out[0], behavior)


evaluate.evaluate = evaluate


def re_evaluate(local_dict=None):
    numexpr = import_numexpr()

    try:
        compiled_ex = numexpr.necompiler._numexpr_last["ex"]  # noqa: F841
    except KeyError:
        raise RuntimeError(
            "not a previous evaluate() execution found"
            + ak._util.exception_suffix(__file__)
        )
    names = numexpr.necompiler._numexpr_last["argnames"]
    arguments = getArguments(names, local_dict)

    arrays = [
        ak.operations.convert.to_layout(x, allow_record=True, allow_other=True)
        for x in arguments
    ]

    def getfunction(inputs, depth):
        if all(
            isinstance(x, ak.layout.NumpyArray) or not isinstance(x, ak.layout.Content)
            for x in inputs
        ):
            return lambda: (
                ak.layout.NumpyArray(numexpr.re_evaluate(dict(zip(names, inputs)))),
            )
        else:
            return None

    behavior = ak._util.behaviorof(*arrays)
    out = ak._util.broadcast_and_apply(
        arrays, getfunction, behavior, allow_records=False
    )
    assert isinstance(out, tuple) and len(out) == 1
    return ak._util.wrap(out[0], behavior)


evaluate.re_evaluate = re_evaluate



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_0166_IndexedOptionArray():
    array = ak.Array([[2, 3, 5], None, [], [7, 11], None, [13], None, [17, 19]])
    assert ak.to_list(ak.prod(array, axis=-1)) == [30, None, 1, 77, None, 13, None, 323]

    array = ak.Array(
        [[[2, 3], [5]], None, [], [[7], [11]], None, [[13]], None, [[17, 19]]]
    )
    assert ak.to_list(ak.prod(array, axis=-1)) == [
        [6, 5],
        None,
        [],
        [7, 11],
        None,
        [13],
        None,
        [323],
    ]

    array = ak.Array([[[2, 3], None, [5]], [], [[7], [11]], [[13]], [None, [17], [19]]])
    ak.to_list(ak.prod(array, axis=-1)) == [
        [6, None, 5],
        [],
        [7, 11],
        [13],
        [None, 17, 19],
    ]

    array = ak.Array([[6, None, 5], [], [7, 11], [13], [None, 17, 19]])
    assert ak.to_list(ak.prod(array, axis=-1)) == [30, 1, 77, 13, 323]


def test_0166_ByteMaskedArray():
    content = ak.from_iter(
        [[2, 3, 5], [999], [], [7, 11], [], [13], [123, 999], [17, 19]], highlevel=False
    )
    mask = ak.layout.Index8(np.array([0, 1, 0, 0, 1, 0, 1, 0], dtype=np.int8))
    array = ak.Array(ak.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert ak.to_list(array) == [
        [2, 3, 5],
        None,
        [],
        [7, 11],
        None,
        [13],
        None,
        [17, 19],
    ]
    assert ak.to_list(ak.prod(array, axis=-1)) == [30, None, 1, 77, None, 13, None, 323]

    content = ak.from_iter(
        [
            [[2, 3], [5]],
            [[999]],
            [],
            [[7], [11]],
            [],
            [[13]],
            [[123], [999]],
            [[17, 19]],
        ],
        highlevel=False,
    )
    mask = ak.layout.Index8(np.array([0, 1, 0, 0, 1, 0, 1, 0], dtype=np.int8))
    array = ak.Array(ak.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert ak.to_list(array) == [
        [[2, 3], [5]],
        None,
        [],
        [[7], [11]],
        None,
        [[13]],
        None,
        [[17, 19]],
    ]
    assert ak.to_list(ak.prod(array, axis=-1)) == [
        [6, 5],
        None,
        [],
        [7, 11],
        None,
        [13],
        None,
        [323],
    ]

    content = ak.from_iter(
        [[2, 3], [999], [5], [7], [11], [13], [], [17], [19]], highlevel=False
    )
    mask = ak.layout.Index8(np.array([0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8))
    bytemasked = ak.layout.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9], dtype=np.int64))
    array = ak.Array(ak.layout.ListOffsetArray64(offsets, bytemasked))
    array = ak.Array([[[2, 3], None, [5]], [], [[7], [11]], [[13]], [None, [17], [19]]])
    assert ak.to_list(ak.prod(array, axis=-1)) == [
        [6, None, 5],
        [],
        [7, 11],
        [13],
        [None, 17, 19],
    ]

    content = ak.from_iter([6, None, 5, 7, 11, 13, None, 17, 19], highlevel=False)
    mask = ak.layout.Index8(np.array([0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=np.int8))
    bytemasked = ak.layout.ByteMaskedArray(mask, content, valid_when=False)
    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 9], dtype=np.int64))
    array = ak.Array(ak.layout.ListOffsetArray64(offsets, bytemasked))
    assert ak.to_list(array) == [[6, None, 5], [], [7, 11], [13], [None, 17, 19]]
    assert ak.to_list(ak.prod(array, axis=-1)) == [30, 1, 77, 13, 323]


def test_0167_strings():
    array = ak.Array(["one", "two", "three", "two", "two", "one", "three"])
    assert ak.to_list(array == "two") == [False, True, False, True, True, False, False]
    assert ak.to_list("two" == array) == [False, True, False, True, True, False, False]
    assert ak.to_list(array == ["two"]) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(["two"] == array) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(array == ak.Array(["two"])) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(ak.Array(["two"]) == array) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]

    array = ak.Array([["one", "two", "three"], [], ["two"], ["two", "one"], ["three"]])
    assert ak.to_list(array == "two") == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list("two" == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(array == ["two"]) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(["two"] == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(array == ak.Array(["two"])) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(ak.Array(["two"]) == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]

    array = ak.Array([["one", "two", "three"], [], ["two"], ["two", "one"], ["three"]])
    assert ak.to_list(array == ["three", "two", "one", "one", "three"]) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list(["three", "two", "one", "one", "three"] == array) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list(array == ak.Array(["three", "two", "one", "one", "three"])) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list(ak.Array(["three", "two", "one", "one", "three"]) == array) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]


def test_0167_bytestrings():
    array = ak.Array([b"one", b"two", b"three", b"two", b"two", b"one", b"three"])
    assert ak.to_list(array == b"two") == [False, True, False, True, True, False, False]
    assert ak.to_list(b"two" == array) == [False, True, False, True, True, False, False]
    assert ak.to_list(array == [b"two"]) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list([b"two"] == array) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(array == ak.Array([b"two"])) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]
    assert ak.to_list(ak.Array([b"two"]) == array) == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
    ]

    array = ak.Array(
        [[b"one", b"two", b"three"], [], [b"two"], [b"two", b"one"], [b"three"]]
    )
    assert ak.to_list(array == b"two") == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(b"two" == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(array == [b"two"]) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list([b"two"] == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(array == ak.Array([b"two"])) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]
    assert ak.to_list(ak.Array([b"two"]) == array) == [
        [False, True, False],
        [],
        [True],
        [True, False],
        [False],
    ]

    array = ak.Array(
        [[b"one", b"two", b"three"], [], [b"two"], [b"two", b"one"], [b"three"]]
    )
    assert ak.to_list(array == [b"three", b"two", b"one", b"one", b"three"]) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list([b"three", b"two", b"one", b"one", b"three"] == array) == [
        [False, False, True],
        [],
        [False],
        [False, True],
        [True],
    ]
    assert ak.to_list(
        array == ak.Array([b"three", b"two", b"one", b"one", b"three"])
    ) == [[False, False, True], [], [False], [False, True], [True]]
    assert ak.to_list(
        ak.Array([b"three", b"two", b"one", b"one", b"three"]) == array
    ) == [[False, False, True], [], [False], [False, True], [True]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_this_should_raise_a_warning():
    ak.deprecations_as_errors = True

    one = ak.Array([{"x": 1}, {"x": 2}, {"x": 3}])
    two = ak.Array([{"x": 1.1}, {"x": 2.2}, {"x": 3.3}])

    with pytest.raises(ValueError):
        one + two


def test_this_should_not():
    def overload_add(left, right):
        return ak.Array({"x": left.x + right.x})

    behavior = {}
    behavior[np.add, "Overload", "Overload"] = overload_add

    one = ak.Array(
        [{"x": 1}, {"x": 2}, {"x": 3}], with_name="Overload", behavior=behavior
    )
    two = ak.Array(
        [{"x": 1.1}, {"x": 2.2}, {"x": 3.3}], with_name="Overload", behavior=behavior
    )

    assert (one + two).tolist() == [{"x": 2.1}, {"x": 4.2}, {"x": 6.3}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert ak.to_list(ak.prod(ak.Array([[[2, 3, 5]], [[7], [11]], [[]]]), axis=-1)) == [
        [30],
        [7, 11],
        [1],
    ]

    assert ak.to_list(ak.prod(ak.Array([[[2, 3, 5]], [[7], [11]], []]), axis=-1)) == [
        [30],
        [7, 11],
        [],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    np_content = np.asfortranarray(np.arange(15).reshape(3, 5))
    ak_content = ak.layout.NumpyArray(np_content)
    offsets = ak.layout.Index64(np.array([0, 0, 1, 1, 2, 2, 3, 3]))
    listoffsetarray = ak.layout.ListOffsetArray64(offsets, ak_content)
    assert ak.to_list(listoffsetarray[1, 0]) == [0, 1, 2, 3, 4]
    assert ak.to_list(listoffsetarray[3, 0]) == [5, 6, 7, 8, 9]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    data = ak.Array([[7, 5, 7], [], [2], [8, 2]])
    assert ak.to_list(ak.sort(data)) == [[5, 7, 7], [], [2], [2, 8]]

    index = ak.argsort(data)
    assert ak.to_list(data[index]) == [[5, 7, 7], [], [2], [2, 8]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

numexpr = pytest.importorskip("numexpr")


def test_numexpr():
    a = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)
    b = ak.Array([100, 200, 300], check_valid=True)
    assert ak.to_list(ak.numexpr.evaluate("a + b")) == [
        [101.1, 102.2, 103.3],
        [],
        [304.4, 305.5],
    ]
    a = [1, 2, 3]
    assert ak.to_list(ak.numexpr.re_evaluate()) == [101, 202, 303]


def test_broadcast_arrays():
    a = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)
    b = ak.Array([100, 200, 300], check_valid=True)

    out = ak.broadcast_arrays(a, b)
    assert ak.to_list(out[0]) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert ak.to_list(out[1]) == [[100, 100, 100], [], [300, 300]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_index():
    array_i1 = np.array([np.iinfo("i1").min, -1, 0, 1, np.iinfo("i1").max], dtype="i1")
    array_u1 = np.array([np.iinfo("u1").min, -1, 0, 1, np.iinfo("u1").max], dtype="u1")
    array_li2 = np.array(
        [np.iinfo("<i2").min, -1, 0, 1, np.iinfo("<i2").max], dtype="<i2"
    )
    array_lu2 = np.array(
        [np.iinfo("<u2").min, -1, 0, 1, np.iinfo("<u2").max], dtype="<u2"
    )
    array_li4 = np.array(
        [np.iinfo("<i4").min, -1, 0, 1, np.iinfo("<i4").max], dtype="<i4"
    )
    array_lu4 = np.array(
        [np.iinfo("<u4").min, -1, 0, 1, np.iinfo("<u4").max], dtype="<u4"
    )
    array_li8 = np.array(
        [np.iinfo("<i8").min, -1, 0, 1, np.iinfo("<i8").max], dtype="<i8"
    )
    array_lu8 = np.array(
        [np.iinfo("<u8").min, -1, 0, 1, np.iinfo("<u8").max], dtype="<u8"
    )
    array_bi2 = np.array(
        [np.iinfo(">i2").min, -1, 0, 1, np.iinfo(">i2").max], dtype=">i2"
    )
    array_bu2 = np.array(
        [np.iinfo(">u2").min, -1, 0, 1, np.iinfo(">u2").max], dtype=">u2"
    )
    array_bi4 = np.array(
        [np.iinfo(">i4").min, -1, 0, 1, np.iinfo(">i4").max], dtype=">i4"
    )
    array_bu4 = np.array(
        [np.iinfo(">u4").min, -1, 0, 1, np.iinfo(">u4").max], dtype=">u4"
    )
    array_bi8 = np.array(
        [np.iinfo(">i8").min, -1, 0, 1, np.iinfo(">i8").max], dtype=">i8"
    )
    array_bu8 = np.array(
        [np.iinfo(">u8").min, -1, 0, 1, np.iinfo(">u8").max], dtype=">u8"
    )

    index_i1 = ak.layout.Index8(array_i1)
    index_u1 = ak.layout.IndexU8(array_u1)
    index_li2 = ak.layout.Index32(array_li2)
    index_lu2 = ak.layout.Index32(array_lu2)
    index_li4 = ak.layout.Index32(array_li4)
    index_lu4 = ak.layout.IndexU32(array_lu4)
    index_li8 = ak.layout.Index64(array_li8)
    index_lu8 = ak.layout.Index64(array_lu8)
    index_bi2 = ak.layout.Index32(array_bi2)
    index_bu2 = ak.layout.Index32(array_bu2)
    index_bi4 = ak.layout.Index32(array_bi4)
    index_bu4 = ak.layout.IndexU32(array_bu4)
    index_bi8 = ak.layout.Index64(array_bi8)
    index_bu8 = ak.layout.Index64(array_bu8)

    assert index_i1[2] == 0
    assert index_u1[2] == 0
    assert index_li2[2] == 0
    assert index_lu2[2] == 0
    assert index_li4[2] == 0
    assert index_lu4[2] == 0
    assert index_li8[2] == 0
    assert index_lu8[2] == 0
    assert index_bi2[2] == 0
    assert index_bu2[2] == 0
    assert index_bi4[2] == 0
    assert index_bu4[2] == 0
    assert index_bi8[2] == 0
    assert index_bu8[2] == 0

    array_i1[2] = 10
    array_u1[2] = 10
    array_li2[2] = 10
    array_lu2[2] = 10
    array_li4[2] = 10
    array_lu4[2] = 10
    array_li8[2] = 10
    array_lu8[2] = 10
    array_bi2[2] = 10
    array_bu2[2] = 10
    array_bi4[2] = 10
    array_bu4[2] = 10
    array_bi8[2] = 10
    array_bu8[2] = 10

    assert index_i1[2] == 10
    assert index_u1[2] == 10
    assert index_li2[2] == 0
    assert index_lu2[2] == 0
    assert index_li4[2] == 10
    assert index_lu4[2] == 10
    assert index_li8[2] == 10
    assert index_lu8[2] == 0
    assert index_bi2[2] == 0
    assert index_bu2[2] == 0
    assert index_bi4[2] == 0
    assert index_bu4[2] == 0
    assert index_bi8[2] == 0
    assert index_bu8[2] == 0


content = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
starts1 = ak.layout.IndexU32(np.array([0, 3, 3, 5, 6], np.uint32))
stops1 = ak.layout.IndexU32(np.array([3, 3, 5, 6, 9], np.uint32))
offsets1 = ak.layout.IndexU32(np.array([0, 3, 3, 5, 6, 9], np.uint32))
starts2 = ak.layout.IndexU32(np.array([0, 2, 3, 3], np.uint32))
stops2 = ak.layout.IndexU32(np.array([2, 3, 3, 5], np.uint32))
offsets2 = ak.layout.IndexU32(np.array([0, 2, 3, 3, 5], np.uint32))


def test_listarray_basic():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    assert ak.to_list(array1) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]
    assert ak.to_list(array1[2]) == [4.4, 5.5]
    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert ak.to_list(array2) == [
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [],
        [[6.6], [7.7, 8.8, 9.9]],
    ]
    assert ak.to_list(array2[1]) == [[4.4, 5.5]]
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]


def test_listoffsetarray_basic():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    assert ak.to_list(array1) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]
    assert ak.to_list(array1[2]) == [4.4, 5.5]
    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert ak.to_list(array2) == [
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [],
        [[6.6], [7.7, 8.8, 9.9]],
    ]
    assert ak.to_list(array2[1]) == [[4.4, 5.5]]
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]


def test_listarray_at():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    assert ak.to_list(array1[2]) == [4.4, 5.5]
    assert ak.to_list(array1[2,]) == [4.4, 5.5]
    assert ak.to_list(array1[2, 1:]) == [5.5]
    assert ak.to_list(array1[2:, 0]) == [4.4, 6.6, 7.7]
    assert ak.to_list(array1[2:, -1]) == [5.5, 6.6, 9.9]


def test_listoffsetarray_at():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    assert ak.to_list(array1[2,]) == [4.4, 5.5]
    assert ak.to_list(array1[2, 1:]) == [5.5]
    assert ak.to_list(array1[2:, 0]) == [4.4, 6.6, 7.7]
    assert ak.to_list(array1[2:, -1]) == [5.5, 6.6, 9.9]


def test_listarray_slice():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert ak.to_list(array1[1:-1,]) == [[], [4.4, 5.5], [6.6]]
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert ak.to_list(array2[1:-1,]) == [[[4.4, 5.5]], []]


def test_listoffsetarray_slice():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert ak.to_list(array1[1:-1,]) == [[], [4.4, 5.5], [6.6]]
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert ak.to_list(array2[1:-1,]) == [[[4.4, 5.5]], []]


def test_listarray_slice_slice():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    assert ak.to_list(array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert ak.to_list(array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert ak.to_list(array1[2:, :-1]) == [[4.4], [], [7.7, 8.8]]


def test_listoffsetarray_slice_slice():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    assert ak.to_list(array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert ak.to_list(array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert ak.to_list(array1[2:, :-1]) == [[4.4], [], [7.7, 8.8]]


def test_listarray_ellipsis():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    if not ak._util.py27:
        assert ak.to_list(array1[Ellipsis, 1:]) == [
            [2.2, 3.3],
            [],
            [5.5],
            [],
            [8.8, 9.9],
        ]
        assert ak.to_list(array2[Ellipsis, 1:]) == [
            [[2.2, 3.3], []],
            [[5.5]],
            [],
            [[], [8.8, 9.9]],
        ]


def test_listoffsetarray_ellipsis():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    if not ak._util.py27:
        assert ak.to_list(array1[Ellipsis, 1:]) == [
            [2.2, 3.3],
            [],
            [5.5],
            [],
            [8.8, 9.9],
        ]
        assert ak.to_list(array2[Ellipsis, 1:]) == [
            [[2.2, 3.3], []],
            [[5.5]],
            [],
            [[], [8.8, 9.9]],
        ]


def test_listarray_array_slice():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0]]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [
        [[2.2, 3.3], []],
        [[2.2, 3.3], []],
        [[5.5]],
        [[5.5]],
        [[5.5]],
        [[2.2, 3.3], []],
    ]


def test_listoffsetarray_array_slice():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0]]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [
        [[2.2, 3.3], []],
        [[2.2, 3.3], []],
        [[5.5]],
        [[5.5]],
        [[5.5]],
        [[2.2, 3.3], []],
    ]


def test_listarray_array():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    assert ak.to_list(array1[np.array([2, 0, 0, 1, -1])]) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        [],
        [7.7, 8.8, 9.9],
    ]
    assert ak.to_list(array1[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])]) == [
        5.5,
        2.2,
        1.1,
        7.7,
    ]

    content_deep = ak.layout.NumpyArray(
        np.array(
            [
                [0, 0],
                [1, 10],
                [2, 20],
                [3, 30],
                [4, 40],
                [5, 50],
                [6, 60],
                [7, 70],
                [8, 80],
            ]
        )
    )
    starts1_deep = ak.layout.IndexU32(np.array([0, 3, 6]))
    stops1_deep = ak.layout.IndexU32(np.array([3, 6, 9]))
    array1_deep = ak.layout.ListArrayU32(starts1_deep, stops1_deep, content_deep)

    assert ak.to_list(array1_deep) == [
        [[0, 0], [1, 10], [2, 20]],
        [[3, 30], [4, 40], [5, 50]],
        [[6, 60], [7, 70], [8, 80]],
    ]
    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), np.array([0, 1, 0, 1]))
    assert np.array(
        [
            [[0, 0], [1, 10], [2, 20]],
            [[3, 30], [4, 40], [5, 50]],
            [[6, 60], [7, 70], [8, 80]],
        ]
    )[s].tolist() == ak.to_list(array1_deep[s])

    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), slice(1, None))
    assert np.array(
        [
            [[0, 0], [1, 10], [2, 20]],
            [[3, 30], [4, 40], [5, 50]],
            [[6, 60], [7, 70], [8, 80]],
        ]
    )[s].tolist() == ak.to_list(array1_deep[s])


def test_listoffsetarray_array():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    assert ak.to_list(array1[np.array([2, 0, 0, 1, -1])]) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        [],
        [7.7, 8.8, 9.9],
    ]
    assert ak.to_list(array1[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])]) == [
        5.5,
        2.2,
        1.1,
        7.7,
    ]

    content_deep = ak.layout.NumpyArray(
        np.array(
            [
                [0, 0],
                [1, 10],
                [2, 20],
                [3, 30],
                [4, 40],
                [5, 50],
                [6, 60],
                [7, 70],
                [8, 80],
            ]
        )
    )
    starts1_deep = ak.layout.IndexU32(np.array([0, 3, 6]))
    stops1_deep = ak.layout.IndexU32(np.array([3, 6, 9]))
    array1_deep = ak.layout.ListArrayU32(starts1_deep, stops1_deep, content_deep)

    assert ak.to_list(array1_deep) == [
        [[0, 0], [1, 10], [2, 20]],
        [[3, 30], [4, 40], [5, 50]],
        [[6, 60], [7, 70], [8, 80]],
    ]
    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), np.array([0, 1, 0, 1]))
    assert np.array(
        [
            [[0, 0], [1, 10], [2, 20]],
            [[3, 30], [4, 40], [5, 50]],
            [[6, 60], [7, 70], [8, 80]],
        ]
    )[s].tolist() == ak.to_list(array1_deep[s])

    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), slice(1, None))
    assert np.array(
        [
            [[0, 0], [1, 10], [2, 20]],
            [[3, 30], [4, 40], [5, 50]],
            [[6, 60], [7, 70], [8, 80]],
        ]
    )[s].tolist() == ak.to_list(array1_deep[s])



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import json

try:
    # pybind11 only supports cPickle protocol 2+ (-1 in pickle.dumps)
    # (automatically satisfied in Python 3; this is just to keep testing Python 2.7)
    import cPickle as pickle
except ImportError:
    import pickle

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_forms():
    form = ak.forms.NumpyForm([], 8, "d")
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert form.inner_shape == []
    assert form.itemsize == 8
    assert form.primitive == "float64"
    assert form.has_identities == False
    assert form.parameters == {}
    assert form.form_key is None
    assert json.loads(form.tojson(False, True)) == {
        "class": "NumpyArray",
        "inner_shape": [],
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
        "has_identities": False,
        "parameters": {},
        "form_key": None,
    }
    assert json.loads(str(form)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }

    form = ak.forms.NumpyForm(
        [1, 2, 3],
        8,
        "d",
        has_identities=True,
        parameters={"hey": ["you", {"there": 3}]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert form.inner_shape == [1, 2, 3]
    assert form.itemsize == 8
    assert form.primitive == "float64"
    assert form.has_identities == True
    assert form.parameters == {"hey": ["you", {"there": 3}]}
    assert form.parameter("hey") == ["you", {"there": 3}]
    assert form.form_key == "yowzers"
    assert json.loads(form.tojson(False, True)) == {
        "class": "NumpyArray",
        "inner_shape": [1, 2, 3],
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
        "has_identities": True,
        "parameters": {"hey": ["you", {"there": 3}]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "NumpyArray",
        "inner_shape": [1, 2, 3],
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
        "has_identities": True,
        "parameters": {"hey": ["you", {"there": 3}]},
        "form_key": "yowzers",
    }

    form = ak.forms.BitMaskedForm(
        "i8",
        ak.forms.NumpyForm([], 8, "d"),
        True,
        False,
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "BitMaskedArray",
        "mask": "i8",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "valid_when": True,
        "lsb_order": False,
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "BitMaskedArray",
        "mask": "i8",
        "content": "float64",
        "valid_when": True,
        "lsb_order": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.mask == "i8"
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.valid_when == True
    assert form.lsb_order == False
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.ByteMaskedForm(
        "i8",
        ak.forms.NumpyForm([], 8, "d"),
        True,
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "ByteMaskedArray",
        "mask": "i8",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "valid_when": True,
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "ByteMaskedArray",
        "mask": "i8",
        "content": "float64",
        "valid_when": True,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.mask == "i8"
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.valid_when == True
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.EmptyForm(parameters={"hey": ["you"]}, form_key="yowzers")
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "EmptyArray",
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "EmptyArray",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.IndexedForm(
        "i64",
        ak.forms.NumpyForm([], 8, "d"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "IndexedArray64",
        "index": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "IndexedArray64",
        "index": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.index == "i64"
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.IndexedOptionForm(
        "i64",
        ak.forms.NumpyForm([], 8, "d"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "IndexedOptionArray64",
        "index": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "IndexedOptionArray64",
        "index": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.index == "i64"
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.ListForm(
        "i64",
        "i64",
        ak.forms.NumpyForm([], 8, "d"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "ListArray64",
        "starts": "i64",
        "stops": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "ListArray64",
        "starts": "i64",
        "stops": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.starts == "i64"
    assert form.stops == "i64"
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.ListOffsetForm(
        "i64",
        ak.forms.NumpyForm([], 8, "d"),
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "ListOffsetArray64",
        "offsets": "i64",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "ListOffsetArray64",
        "offsets": "i64",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.offsets == "i64"
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.RecordForm(
        {"one": ak.forms.NumpyForm([], 8, "d"), "two": ak.forms.NumpyForm([], 1, "?")},
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "RecordArray",
        "contents": {
            "one": {
                "class": "NumpyArray",
                "inner_shape": [],
                "itemsize": 8,
                "format": "d",
                "primitive": "float64",
                "has_identities": False,
                "parameters": {},
                "form_key": None,
            },
            "two": {
                "class": "NumpyArray",
                "inner_shape": [],
                "itemsize": 1,
                "format": "?",
                "primitive": "bool",
                "has_identities": False,
                "parameters": {},
                "form_key": None,
            },
        },
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "RecordArray",
        "contents": {"one": "float64", "two": "bool"},
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    if not ak._util.py27 and not ak._util.py35:
        assert [json.loads(str(x)) for x in form.values()] == [
            {
                "class": "NumpyArray",
                "itemsize": 8,
                "format": "d",
                "primitive": "float64",
            },
            {"class": "NumpyArray", "itemsize": 1, "format": "?", "primitive": "bool"},
        ]
        assert {n: json.loads(str(x)) for n, x in form.contents.items()} == {
            "one": {
                "class": "NumpyArray",
                "itemsize": 8,
                "format": "d",
                "primitive": "float64",
            },
            "two": {
                "class": "NumpyArray",
                "itemsize": 1,
                "format": "?",
                "primitive": "bool",
            },
        }
    assert json.loads(str(form.content("one"))) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert json.loads(str(form.content("two"))) == {
        "class": "NumpyArray",
        "itemsize": 1,
        "format": "?",
        "primitive": "bool",
    }
    if not ak._util.py27 and not ak._util.py35:
        assert json.loads(str(form.content(0))) == {
            "class": "NumpyArray",
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
        }
        assert json.loads(str(form.content(1))) == {
            "class": "NumpyArray",
            "itemsize": 1,
            "format": "?",
            "primitive": "bool",
        }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.RecordForm(
        [ak.forms.NumpyForm([], 8, "d"), ak.forms.NumpyForm([], 1, "?")],
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "RecordArray",
        "contents": [
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "itemsize": 8,
                "format": "d",
                "primitive": "float64",
                "has_identities": False,
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "itemsize": 1,
                "format": "?",
                "primitive": "bool",
                "has_identities": False,
                "parameters": {},
                "form_key": None,
            },
        ],
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "RecordArray",
        "contents": ["float64", "bool"],
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert [json.loads(str(x)) for x in form.values()] == [
        {"class": "NumpyArray", "itemsize": 8, "format": "d", "primitive": "float64"},
        {"class": "NumpyArray", "itemsize": 1, "format": "?", "primitive": "bool"},
    ]
    assert {n: json.loads(str(x)) for n, x in form.contents.items()} == {
        "0": {
            "class": "NumpyArray",
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
        },
        "1": {"class": "NumpyArray", "itemsize": 1, "format": "?", "primitive": "bool"},
    }
    assert json.loads(str(form.content(0))) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert json.loads(str(form.content(1))) == {
        "class": "NumpyArray",
        "itemsize": 1,
        "format": "?",
        "primitive": "bool",
    }
    assert json.loads(str(form.content("0"))) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert json.loads(str(form.content("1"))) == {
        "class": "NumpyArray",
        "itemsize": 1,
        "format": "?",
        "primitive": "bool",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.RegularForm(
        ak.forms.NumpyForm([], 8, "d"),
        10,
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "RegularArray",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "size": 10,
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "RegularArray",
        "content": "float64",
        "size": 10,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.size == 10
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.UnionForm(
        "i8",
        "i64",
        [ak.forms.NumpyForm([], 8, "d"), ak.forms.NumpyForm([], 1, "?")],
        parameters={"hey": ["you"]},
        form_key="yowzers",
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "UnionArray8_64",
        "tags": "i8",
        "index": "i64",
        "contents": [
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "itemsize": 8,
                "format": "d",
                "primitive": "float64",
                "has_identities": False,
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "inner_shape": [],
                "itemsize": 1,
                "format": "?",
                "primitive": "bool",
                "has_identities": False,
                "parameters": {},
                "form_key": None,
            },
        ],
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "UnionArray8_64",
        "tags": "i8",
        "index": "i64",
        "contents": ["float64", "bool"],
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert form.tags == "i8"
    assert form.index == "i64"
    assert json.loads(str(form.contents)) == [
        {"class": "NumpyArray", "itemsize": 8, "format": "d", "primitive": "float64"},
        {"class": "NumpyArray", "itemsize": 1, "format": "?", "primitive": "bool"},
    ]
    assert json.loads(str(form.content(0))) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert json.loads(str(form.content(1))) == {
        "class": "NumpyArray",
        "itemsize": 1,
        "format": "?",
        "primitive": "bool",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.UnmaskedForm(
        ak.forms.NumpyForm([], 8, "d"), parameters={"hey": ["you"]}, form_key="yowzers"
    )
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert json.loads(form.tojson(False, True)) == {
        "class": "UnmaskedArray",
        "content": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "has_identities": False,
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form)) == {
        "class": "UnmaskedArray",
        "content": "float64",
        "parameters": {"hey": ["you"]},
        "form_key": "yowzers",
    }
    assert json.loads(str(form.content)) == {
        "class": "NumpyArray",
        "itemsize": 8,
        "format": "d",
        "primitive": "float64",
    }
    assert form.has_identities == False
    assert form.parameters == {"hey": ["you"]}
    assert form.parameter("hey") == ["you"]
    assert form.form_key == "yowzers"

    form = ak.forms.VirtualForm(ak.forms.NumpyForm([], 8, "d"), True)
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert form.form.inner_shape == []
    assert form.form.itemsize == 8
    assert form.form.primitive == "float64"
    assert form.form.has_identities == False
    assert form.form.parameters == {}
    assert form.has_length is True
    assert form.parameters == {}
    assert json.loads(form.tojson(False, True)) == {
        "class": "VirtualArray",
        "form": {
            "class": "NumpyArray",
            "inner_shape": [],
            "itemsize": 8,
            "format": "d",
            "primitive": "float64",
            "has_identities": False,
            "parameters": {},
            "form_key": None,
        },
        "has_length": True,
        "has_identities": False,
        "parameters": {},
        "form_key": None,
    }
    assert json.loads(str(form)) == {
        "class": "VirtualArray",
        "form": "float64",
        "has_length": True,
    }

    form = ak.forms.VirtualForm(None, False)
    assert form == form
    assert pickle.loads(pickle.dumps(form, -1)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, False)) == form
    assert ak.forms.Form.fromjson(form.tojson(False, True)) == form
    assert form.form is None
    assert form.has_length is False
    assert form.parameters == {}
    assert json.loads(form.tojson(False, True)) == {
        "class": "VirtualArray",
        "form": None,
        "has_length": False,
        "has_identities": False,
        "parameters": {},
        "form_key": None,
    }
    assert json.loads(str(form)) == {
        "class": "VirtualArray",
        "form": None,
        "has_length": False,
    }


def fcn():
    return ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5]))


def test_basic():
    generator = ak.layout.ArrayGenerator(
        fcn, form=ak.forms.NumpyForm([], 8, "d"), length=5
    )

    d = ak._util.MappingProxy({})
    cache = ak.layout.ArrayCache(d)

    virtualarray = ak.layout.VirtualArray(generator, cache)
    assert virtualarray.peek_array is None
    assert virtualarray.array is not None
    assert ak.to_list(virtualarray.peek_array) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert ak.to_list(virtualarray.array) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert ak.to_list(d[virtualarray.cache_key]) == [1.1, 2.2, 3.3, 4.4, 5.5]

    cache = ak.layout.ArrayCache(None)

    virtualarray = ak.layout.VirtualArray(generator, cache)
    assert virtualarray.peek_array is None
    assert virtualarray.array is not None
    assert virtualarray.peek_array is None
    assert ak.to_list(virtualarray.array) == [1.1, 2.2, 3.3, 4.4, 5.5]


def test_slice():
    generator = ak.layout.ArrayGenerator(
        lambda: ak.Array(
            [[1.1, 2.2, 3.3, 4.4, 5.5], [6.6, 7.7, 8.8], [100, 200, 300, 400]]
        ),
        length=3,
    )
    virtualarray = ak.layout.VirtualArray(generator)

    assert isinstance(virtualarray, ak.layout.VirtualArray)

    sliced = virtualarray[:-1]
    assert isinstance(sliced, ak.layout.VirtualArray)

    assert isinstance(sliced[1], ak.layout.NumpyArray)


def test_field():
    generator = ak.layout.ArrayGenerator(
        lambda: ak.Array(
            [
                {"x": 0.0, "y": []},
                {"x": 1.1, "y": [1]},
                {"x": 2.2, "y": [2, 2]},
                {"x": 3.3, "y": [3, 3, 3]},
            ]
        )
    )
    virtualarray = ak.layout.VirtualArray(generator)

    assert isinstance(virtualarray, ak.layout.VirtualArray)

    sliced = virtualarray["y"]
    assert isinstance(sliced, ak.layout.VirtualArray)

    assert isinstance(sliced[1], ak.layout.NumpyArray)


def test_single_level():
    template = ak.Array(
        [
            [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}],
            [],
            [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}],
        ]
    )
    generator = ak.layout.ArrayGenerator(
        lambda: template, form=template.layout.form, length=3
    )
    d = ak._util.MappingProxy({})
    cache = ak.layout.ArrayCache(d)
    virtualarray = ak.layout.VirtualArray(generator, cache)

    a = virtualarray[2]
    assert isinstance(a, ak.layout.RecordArray)
    assert len(d) == 1
    assert ak.to_list(a) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    d.clear()

    a = virtualarray[1:]
    assert isinstance(a, ak.layout.VirtualArray)
    assert len(d) == 0
    b = a[1]
    assert isinstance(b, ak.layout.RecordArray)
    assert len(d) == 1
    assert ak.to_list(b) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    d.clear()

    a = virtualarray[[0, 2, 1, 0]]
    assert isinstance(a, ak.layout.VirtualArray)
    assert len(d) == 0
    b = a[1]
    assert isinstance(b, ak.layout.RecordArray)
    assert len(d) == 1
    assert ak.to_list(b) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    d.clear()

    a = virtualarray[[False, True, True]]
    assert isinstance(a, ak.layout.VirtualArray)
    assert len(d) == 0
    b = a[1]
    assert isinstance(b, ak.layout.RecordArray)
    assert len(d) == 1
    assert ak.to_list(b) == [{"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    d.clear()

    a = virtualarray["x"]
    assert isinstance(a, ak.layout.VirtualArray)
    assert len(d) == 0
    b = a[2]
    assert isinstance(b, ak.layout.NumpyArray)
    assert len(d) == 1
    assert ak.to_list(b) == [3.3, 4.4]
    d.clear()

    a = virtualarray["y"]
    assert isinstance(a, ak.layout.VirtualArray)
    assert len(d) == 0
    b = a[2]
    assert isinstance(b, (ak.layout.ListArray64, ak.layout.ListOffsetArray64))
    assert len(d) == 1
    assert ak.to_list(b) == [[3, 3, 3], [4, 4, 4, 4]]
    d.clear()

    a = virtualarray[::2, 1]
    assert isinstance(a, (ak.layout.RecordArray, ak.layout.IndexedArray64))
    assert len(d) == 1
    assert ak.to_list(a) == [{"x": 1.1, "y": [1]}, {"x": 4.4, "y": [4, 4, 4, 4]}]
    d.clear()


def test_iter():
    generator = ak.layout.ArrayGenerator(
        fcn, form=ak.forms.NumpyForm([], 8, "d"), length=5
    )
    d = ak._util.MappingProxy({})
    cache = ak.layout.ArrayCache(d)
    virtualarray = ak.layout.VirtualArray(generator, cache)

    assert len(d) == 0
    it = iter(virtualarray)
    assert len(d) == 1
    d.clear()
    assert len(d) == 0
    assert next(it) == 1.1
    assert len(d) == 0
    assert list(it) == [2.2, 3.3, 4.4, 5.5]
    assert len(d) == 0


def test_nested_virtualness():
    counter = [0, 0]

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )

    def materialize1():
        counter[1] += 1
        return content

    generator1 = ak.layout.ArrayGenerator(
        materialize1, form=content.form, length=len(content)
    )
    virtual1 = ak.layout.VirtualArray(generator1)

    offsets = ak.layout.Index64(np.array([0, 3, 3, 5, 6, 10], dtype=np.int64))
    listarray = ak.layout.ListOffsetArray64(offsets, virtual1)

    def materialize2():
        counter[0] += 1
        return listarray

    generator2 = ak.layout.ArrayGenerator(
        materialize2, form=listarray.form, length=len(listarray)
    )
    virtual2 = ak.layout.VirtualArray(generator2)

    assert counter == [0, 0]

    tmp1 = virtual2[2]
    assert isinstance(tmp1, ak.layout.VirtualArray)
    assert counter == [1, 0]

    tmp2 = tmp1[1]
    assert tmp2 == 4.4
    assert counter == [1, 1]


def test_highlevel():
    array = ak.virtual(lambda: [[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    assert isinstance(array.layout, ak.layout.VirtualArray)
    assert ak.to_list(array) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    counter = [0]

    def generate():
        counter[0] += 1
        return [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    array = ak.virtual(
        generate,
        length=3,
        form={"class": "ListOffsetArray64", "offsets": "i64", "content": "float64"},
    )
    assert counter[0] == 0

    assert len(array) == 3
    assert counter[0] == 0

    assert str(ak.type(array)) == "3 * var * float64"
    assert counter[0] == 0

    assert ak.to_list(array[2]) == [4.4, 5.5]
    assert counter[0] == 1



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_broadcast_single_bool():
    base = ak.Array([[{"x": 0.1, "y": 0.2, "z": 0.3}, {"x": 0.4, "y": 0.5, "z": 0.6}]])
    base_new1 = ak.with_field(base, True, "always_true")
    assert ak.to_list(base_new1.always_true) == [[True, True]]
    base_new2 = ak.with_field(base_new1, base.x > 0.3, "sometimes_true")
    assert ak.to_list(base_new2.always_true) == [[True, True]]
